import 'dart:convert';
import 'dart:io';
import 'package:path/path.dart';
import 'package:recase/recase.dart';
import '../lib/craftsman.dart' show Utils;

main(List<String> args) {
  //handle command
  _scaffold();
}

void _scaffold() {
  if (!Directory("craftsman/descriptors").existsSync()) {
    Utils.log("Directory 'craftsman/descriptors' does not exists", color: 'red');
    return Utils.log("Run 'craftsman:descriptor' command to create a new descriptor template, then compile and run 'craftsman:scaffold' again", color: 'yellow');
  }

  Utils.log("Let's iterate 'craftsman/descriptors' directory", color: 'green');
  final directory = Directory("craftsman/descriptors");
  final contents = directory.listSync();

  if (contents.length <= 0) {
    return Utils.log("No descriptors found in 'craftsman/descriptors' directory", color: 'red');
  }

  // last version
  int lastVersion = 1;

  for (var fileOrDir in contents) {
    if (fileOrDir is File) {
      // check version
      if (basename(fileOrDir.path) == '.craft') {
        try {
          final fileContent = fileOrDir.readAsStringSync();
          final fileContentDecoded = json.decode(fileContent) as Map<String, dynamic>;
          lastVersion = fileContentDecoded['LAST_VERSION'];
        } catch (e) {}
        continue;
      }

      final fileContent = fileOrDir.readAsStringSync();
      try {
        final fileContentDecoded = json.decode(fileContent) as Map<String, dynamic>;

        if (!fileContentDecoded.containsKey('db')) return Utils.log("Descriptors must have a root 'db' key for DataBase definition", color: 'red');
        if (!fileContentDecoded.containsKey('orm')) return Utils.log("Descriptors must have a root 'orm' key for Object Relational Mapper definition", color: 'red');

        final descriptorVersion = fileContentDecoded['db']['version'];
        if (descriptorVersion <= lastVersion) {
          Utils.log("Skipping '${fileOrDir.path}' descriptor file, the descriptor version ($descriptorVersion) is less or equal than the last scaffolded version ($lastVersion)", color: 'blue');
          continue;
        }
        Utils.log("Reading  '${fileOrDir.path}' descriptor file", color: 'green');

        // generate strings to write on files
        final migrationsToWrite = _createMigrations(fileContentDecoded['db']);
        final modelsToWrite = _createModels(fileContentDecoded['orm'], fileContentDecoded['db']);

        // get db version
        final dbVersion = fileContentDecoded['db']['version'];

        // write migrations to file
        final migrationsPath = "assets/loquacious/migrations/$dbVersion.json";
        bool writeMigration = true;
        if (File(migrationsPath).existsSync()) {
          Utils.log("Migration file with name '$dbVersion.json' already exists, do you want to overwrite? (y = yes | n = no)", color: 'yellow');
          var line = stdin.readLineSync(encoding: Encoding.getByName('utf-8'));
          if (line.trim().toLowerCase() != 'y' && line.trim().toLowerCase() != 'yes') {
            Utils.log("Skipping scaffolding build for '$migrationsPath'", color: 'blue');
            writeMigration = false;
          }
        }
        if (writeMigration) {
          new File(migrationsPath).create(recursive: true).then((file) {
            file.writeAsStringSync(migrationsToWrite);
          });
          Utils.log("Created '$migrationsPath' migration file", color: 'green');
        }

        // write models to file
        final modelsPath = "lib/models";
        for (var entityName in modelsToWrite.keys) {
          final modelPath = '$modelsPath/${entityName.toLowerCase()}.dart';
          bool writeModel = true;
          if (File(modelPath).existsSync()) {
            Utils.log("Model file with name '${entityName.toLowerCase()}.dart' already exists, do you want to overwrite? (y = yes | n = no)", color: 'yellow');
            var line = stdin.readLineSync(encoding: Encoding.getByName('utf-8'));
            if (line.trim().toLowerCase() != 'y' && line.trim().toLowerCase() != 'yes') {
              Utils.log("Skipping scaffolding build for '$modelPath'", color: 'blue');
              writeModel = false;
            }
          }
          if (writeModel) {
            new File(modelPath).create(recursive: true).then((file) {
              file.writeAsStringSync(modelsToWrite[entityName]);
            });
            Utils.log("Created '$modelPath' model file", color: 'green');
          }
        }

        // set last descriptor version
        lastVersion = descriptorVersion;
      } catch (e) {
        Utils.log("Unable to decode '${fileOrDir.path}' descriptor file", color: 'red');
        Utils.log(e.toString(), color: 'red');
      }
    }
  }

  // create a .craft file to store current version
  new File("craftsman/descriptors/.craft").create(recursive: true).then((file) {
    file.writeAsStringSync('{"LAST_VERSION": $lastVersion}');
  });
}

String _createMigrations(Map<String, dynamic> dbData) {
  final List<String> upQueries = [];
  final List<String> downQueries = [];
  for (Map<String, dynamic> migration in dbData['migrations']) {
    if (migration.containsKey('create_table')) {
      final schema = migration['create_table'];
      final tableName = schema['name'];
      final tableColumns = schema['columns'];
      final upQuery = _migrationCreateTable(tableName, tableColumns);
      upQueries.add(upQuery);
      downQueries.add("DROP TABLE IF EXISTS $tableName;");
    }
  }

  // return encoded map
  return json.encode({
    "up": upQueries,
    "down": downQueries,
  });
}

String _migrationCreateTable(String tableName, List<dynamic> tableColumns) {
  String query = "CREATE TABLE IF NOT EXISTS $tableName (";
  List<String> queryParts = [];
  List<String> foreigns = [];
  for (var column in tableColumns) {
    if (column is Map<String, dynamic>) {
      final colName = column['name'];
      final colType = (column['type'] as String).toUpperCase();
      String colQuery = '$colName $colType';

      if (!column.containsKey('nullable') || !column['nullable']) {
        colQuery += ' NOT NULL';
      }

      if (column.containsKey('default')) {
        colQuery += ' DEFAULT ' + column['default'];
      }

      if (column.containsKey('index')) {
        switch (column['index']) {
          case 'primary':
            colQuery += ' PRIMARY KEY';
            break;
          case 'unique':
            colQuery += ' UNIQUE';
            break;
        }
      }

      if (column.containsKey('foreign')) {
        final Map<String, dynamic> foreign = column['foreign'];
        final foreignTable = foreign['table'];
        final foreignColumn = foreign['column'];
        String foreignQuery = "FOREIGN KEY ($colName) REFERENCES $foreignTable ($foreignColumn)";
        if (foreign.containsKey('on_update')) {
          foreignQuery += " ON UPDATE " + (foreign['on_update'] as String).toUpperCase();
        }
        if (foreign.containsKey('on_delete')) {
          foreignQuery += " ON UPDATE " + (foreign['on_delete'] as String).toUpperCase();
        }
        foreigns.add(foreignQuery);
      }

      queryParts.add(colQuery);
    }

    if (column is String) {
      switch (column) {
        case "timestamps":
          queryParts.add("created_at TEXT NOT NULL");
          queryParts.add("updated_at TEXT NOT NULL");
          break;
      }
    }
  }

  query += queryParts.join(', ');
  if (foreigns.length > 0) query += ', ' + foreigns.join(', ');
  query += ');';
  return query;
}

Map<String, String> _createModels(Map<String, dynamic> ormData, Map<String, dynamic> dbData) {
  Map<String, String> returnModels = {};
  if (ormData.containsKey('create_models')) {
    for (var i = 0; i < ormData['create_models'].length; i++) {
      final Map<String, dynamic> modelData = ormData['create_models'][i];
      String entity = (modelData['entity'] as String).pascalCase;
      String table;
      String primary;
      Map<String, dynamic> columns = {};
      bool timestamps;
      if (modelData.containsKey('from_migration_table')) {
        table = modelData['from_migration_table'];
        final dbDataOfTable = dbData['migrations'].where((migration) {
          if (migration.containsKey('create_table')) {
            return table == migration['create_table']['name'];
          }
          return false;
        }).toList()[0];

        dbDataOfTable['create_table']['columns'].forEach((column) {
          if (column is String) {
            if (column == 'timestamps') timestamps = true;
          } else {
            if (column.containsKey('index') && column['index'].toLowerCase() == 'primary') {
              primary = column['name'];
            }
            columns[column['name']] = column['type'];
          }
        });
      } else {
        table = modelData['table'];
        primary = modelData['primary'];
        columns = modelData['columns'];
        timestamps = modelData.containsKey('timestamps') ? modelData['timestamps'] == true : false;
      }

      // read the model template
      String modelTemplate = _getModelTemplate();

      LineSplitter ls = new LineSplitter();
      List<String> lines = ls.convert(modelTemplate);
      List<String> resultLines = [];
      bool getLine = true;
      lines.forEach((line) {
        if (line.contains('#if')) {
          if (line.contains('timestamps') && !timestamps) getLine = false;
        } else if (line.contains('#endif')) {
          getLine = true;
        } else {
          if (getLine) {
            /* 
              TEMPLATE KEYS
              #entity => the model class name
              #table => the table name
              #primary => the primary key column name
              #primary_type => the primary key column type
              #arguments_fields => the fields inside the "arguments" map
              #arguments_props => the arguments props inside the model
              #arguments_constructor_parameters => the arguments as parameters for "constructor" method
              #arguments_constructor_body => the argument assignations into "constructor" method body
              #arguments_map_method => the arguments listed into "mapToModel" method
              #arguments_create_method_parameters => the arguments as parameters for "create" method
              #arguments_create_method_body => the argument assignations into "create" method body
            */

            final regExp = RegExp(r'#entity|#table|#primary\b|#primary_type|#arguments_fields|#arguments_props|#arguments_constructor_parameters|#arguments_constructor_body|#arguments_map_method|#arguments_create_method_parameters|#arguments_create_method_body');
            final resultLine = line.replaceAllMapped(regExp, (match) {
              List<String> c = [];
              switch (match[0]) {
                case '#entity':
                  return entity;
                  break;
                case '#table':
                  return table;
                  break;
                case '#primary':
                  return primary;
                  break;
                case '#primary_type':
                  return _dbTypeToDartType(columns[primary]);
                  break;
                case '#arguments_fields':
                  columns.forEach((name, type) => name == primary ? null : c.add("'$name': null,"));
                  return c.join(' ');
                  break;
                case '#arguments_props':
                  columns.forEach((name, type) => name == primary ? null : c.add("""// $name property
  ${_dbTypeToDartType(type)} get ${name.camelCase} => arguments['$name'];
  set ${name.camelCase}(${_dbTypeToDartType(type)} value) {
    arguments['$name'] = value;
  }"""));
                  return c.join('\n\n  ');
                  break;
                case '#arguments_constructor_parameters':
                case '#arguments_create_method_parameters':
                  columns.forEach((name, type) => name == primary ? null : c.add("${_dbTypeToDartType(type)} ${name.camelCase},"));
                  return c.join(' ');
                  break;
                case '#arguments_constructor_body':
                  columns.forEach((name, type) => name == primary ? null : c.add("this.${name.camelCase} = ${name.camelCase};"));
                  return c.join(' ');
                  break;
                case '#arguments_map_method':
                  columns.forEach((name, type) => name == primary ? null : c.add("${name.camelCase}: el['$name'],"));
                  return c.join(' ');
                  break;
                case '#arguments_create_method_body':
                  columns.forEach((name, type) => name == primary ? null : c.add("m.${name.camelCase} = ${name.camelCase};"));
                  return c.join(' ');
                  break;
              }
              return match[0];
            });

            // add result line
            resultLines.add(resultLine);
          }
        }
      });
      String resultString = resultLines.join('\n');
      returnModels[entity] = resultString;
    }
  }

  return returnModels;
}

String _getModelTemplate() {
  return """/*
  THIS CLASS IS AUTOGENERATED
  You shouldn't modify it manually 
  For creating or modifying models you can create a descriptor and run craftman:scaffold command
 */
import 'package:intl/intl.dart';
import 'package:loquacious/loquacious.dart';

/* 
  The #entity model class
 */
class #entity {
  /* 
    BASE PROPERTIES
   */

  // table name
  String _table = '#table';
  String get table => _table;

  // primary key of the table
  static const String _primaryKey = '#primary';

  // arguments of the model representing the table columns
  Map<String, dynamic> arguments = {
    '#primary': null,
    #arguments_fields
    #if timestamps 
    'created_at': null,
    'updated_at': null,
    #endif
  };

  /* 
    ARGUMENTS PROPERTIES
   */
    
  // primary key property
  #primary_type get #primary => arguments['#primary'];
  set #primary(#primary_type value) {
    arguments['#primary'] = value;
  }

  #arguments_props

  #if timestamps 
  // created at timestamp property
  DateTime get createdAt => DateTime.parse(arguments['created_at']);
  set createdAt(DateTime value) {
    arguments['created_at'] = DateFormat('yyy-MM-dd H:m:s').format(value);
  }

  // updated at timestamp property
  DateTime get updatedAt => DateTime.parse(arguments['updated_at']);
  set updatedAt(DateTime value) {
    arguments['updated_at'] = DateFormat('yyy-MM-dd H:m:s').format(value);
  }
  #endif

  /* 
    CONSTRUCTOR
   */
  #entity({
    #primary_type #primary,
    #arguments_constructor_parameters
  }) {
    this.#primary = #primary;
    #arguments_constructor_body
  }

  /* 
    UTILITY METHODS
   */

  // MAP METHOD
  static #entity mapToModel(dynamic e) {
    final el = Map<String, dynamic>.from(e);
    final m = #entity(
      #primary: el['#primary'],
      #arguments_map_method
    );
    #if timestamps
    m.arguments['created_at'] = el['created_at'];
    m.arguments['updated_at'] = el['updated_at'];
    #endif
    return m;
  }

  /* 
    CREATE
    instantiate a new #entity, save it do DB and returns it
   */
  static Future<#entity> create({
    #primary_type #primary, 
    #arguments_create_method_parameters
  }) async {
    final m = #entity();
    m.#primary = #primary;
    #arguments_create_method_body
    await m.save();
    return m;
  }

  /* 
    ALL
    returns the whole collection of #entity model
   */
  static Future<List<#entity>> all() async {
    return _LQBM<#entity>.table('#table').get();
  }

  /* 
    FIND
    find a #entity model and returns it if found, else return null
   */
  static Future<#entity> find(#primary_type #primary) async {
    return await _LQBM<#entity>.table('#table').where(#entity._primaryKey, #primary).first();
  }

  /* 
    WHERE
    instantiate a Loquacious query builder of type #entity, applies a where and returns it
   */
  static _LQBM<#entity> where(
    String column,
    dynamic value, {
    String comparisonOperator,
  }) {
    return _LQBM<#entity>.table('#table').where(
      column,
      value,
      comparisonOperator: comparisonOperator,
    );
  }

  /* 
    ORDER BY
    instantiate a Loquacious query builder of type #entity, applies an order by and returns it
   */
  static _LQBM<#entity> orderBy(String column) {
    return _LQBM<#entity>.table('#table').orderBy(column);
  }

  /* 
    ORDER BY DESC
    instantiate a Loquacious query builder of type #entity, applies an order by desc and returns it
   */
  static _LQBM<#entity> orderByDesc(String column) {
    return _LQBM<#entity>.table('#table').orderByDesc(column);
  }

  /* 
    GROUP BY
    instantiate a Loquacious query builder of type #entity, applies a group by and returns it
   */
  static _LQBM<#entity> groupBy(List<String> columns) {
    return _LQBM<#entity>.table('#table').groupBy(columns);
  }

  /* 
    LIMIT
    instantiate a Loquacious query builder of type #entity, applies a limit and returns it
   */
  static _LQBM<#entity> limit(int count) {
    return _LQBM<#entity>.table('#table').limit(count);
  }

  /* 
    TAKE
    instantiate a Loquacious query builder of type #entity, applies a take and returns it
   */
  static _LQBM<#entity> take(int count) {
    return _LQBM<#entity>.table('#table').take(count);
  }

  /* 
    OFFSET
    instantiate a Loquacious query builder of type #entity, applies an offset and returns it
   */
  static _LQBM<#entity> offset(int count) {
    return _LQBM<#entity>.table('#table').offset(count);
  }

  /* 
    SKIP
    instantiate a Loquacious query builder of type #entity, applies a skip and returns it
   */
  static _LQBM<#entity> skip(int count) {
    return _LQBM<#entity>.table('#table').skip(count);
  }

  /* 
    SAVE
    save the #entity model to DB
   */
  Future<#entity> save() async {
    #if timestamps
    if (arguments.containsKey('created_at') && arguments['created_at'] == null) {
      createdAt = DateTime.now();
    }
    updatedAt = DateTime.now();
    #endif

    if (arguments.containsKey(_primaryKey) && arguments[_primaryKey] == null) {
      arguments.removeWhere((key, value) => key == _primaryKey);
      #primary_type #primary = await LQB.table(_table).insertGetId(arguments);
      arguments['#primary'] = #primary;
    } else {
      await LQB.table(_table).where(_primaryKey, arguments['#primary']).update(arguments);
    }
    return this;
  }

  @override
  String toString() {
    return arguments.toString();
  }
}

/* 
 QUERY BUILDER OF #entity MODEL
 */
class _LQBM<T extends #entity> extends LQB {
  _LQBM.table(String tableName) : super.table(tableName);

  @override
  Future<List<#entity>> get() async {
    return List<#entity>.from((await this.getDynamic()).map(#entity.mapToModel));
  }

  @override
  Future<#entity> first() async {
    try {
      return List<#entity>.from((await this.getDynamic()).map(#entity.mapToModel)).first;
    } catch (e) {}
    return null;
  }
}""";
}

String _dbTypeToDartType(String dbType) {
  switch (dbType.toLowerCase()) {
    case 'integer':
      return 'int';
      break;
    case 'text':
      return 'String';
      break;
    case 'real':
      return 'num';
      break;
    case 'blob':
      return 'Uint8List';
      break;
    default:
      return 'String';
  }
}
